# EasyLiteQuery.py
import sqlite3
from typing import List, Tuple, Any
from .EasyLiteResult import EasyLiteResult

# Class for building SELECT queries
class EasyLiteQuery:
    # Constructor
    def __init__(self, connection: sqlite3.Connection, table_name: str):
        self.connection = connection
        self.table_name = table_name
        self._fields = ["*"]
        self._where_clauses = []
        self._group_by_columns = []
        self._order_clause = None
        self._limit_count = None
        self._params = []
        self._joins = []

    # Select specific fields
    def fields(self, *fields: str):
        if fields:
            self._fields = list(fields)
        return self

    # Add a WHERE clause
    def where(self, clause: str, *params):
        self._where_clauses.append(clause)
        self._params.extend(params)
        return self

    # Fluent join by referencing a local FK and a remote PK
    def join(self, local_field: str, target_table: str, target_pk: str = "id", join_type: str = "INNER"):
        condition = f"{self.table_name}.{local_field} = {target_table}.{target_pk}"
        self._joins.append((target_table, condition, join_type.upper()))
        return self

    # Custom join for arbitrary ON conditions
    def customJoin(self, other_table: str, on_condition: str, join_type: str = "INNER"):
        self._joins.append((other_table, on_condition, join_type.upper()))
        return self

    # Add a GROUP BY clause
    def groupBy(self, *columns: str):
        for c in columns:
            self._group_by_columns.append(c)
        return self

    # Add an ORDER BY clause
    def sortBy(self, column_name: str, ascending: bool = True):
        direction = "ASC" if ascending else "DESC"
        self._order_clause = (column_name, direction)
        return self

    # Add a LIMIT
    def limit(self, count: int):
        self._limit_count = count
        return self

    # Execute and return results
    def fetch(self) -> EasyLiteResult:
        sql, params = self._build_sql()
        try:
            c = self.connection.cursor()
            c.execute(sql, params)
            rows = c.fetchall()
            cols = [d[0] for d in c.description] if c.description else []
            print('[SUCCESS] Query executed, EasyLiteResult object returned.')
            return EasyLiteResult(rows, cols)
        except sqlite3.Error as e:
            print(f"[ERROR] Failed to execute SELECT query on '{self.table_name}': {e}")
            return EasyLiteResult([], [])

    # Build the final SQL query
    def _build_sql(self) -> Tuple[str, list]:
        fstr = ", ".join(self._fields)
        sql = f"SELECT {fstr} FROM {self.table_name}"

        # Build joins
        for (tbl, cond, jtype) in self._joins:
            sql += f" {jtype} JOIN {tbl} ON {cond}"

        # WHERE
        if self._where_clauses:
            w = " AND ".join(self._where_clauses)
            sql += f" WHERE {w}"

        # GROUP BY
        if self._group_by_columns:
            g = ", ".join(self._group_by_columns)
            sql += f" GROUP BY {g}"

        # ORDER BY
        if self._order_clause:
            col, direct = self._order_clause
            sql += f" ORDER BY {col} {direct}"

        # LIMIT
        if self._limit_count is not None:
            sql += f" LIMIT {self._limit_count}"

        return sql, self._params



# EasyLiteBuild.py
import sqlite3

# Class for building or modifying tables
class EasyLiteBuild:
    # Constructor
    def __init__(self, connection: sqlite3.Connection, table_name: str, mode: str):
        self.connection = connection
        self.table_name = table_name
        self.mode = mode
        self._cols_def = []
        self._fks_def = []
        self._pk_defined = False
        self._col_to_modify = None

    # Create a PK column
    def PK(self, name="id", autoincrement=True):
        if self._pk_defined:
            return self
        pk = f"{name} INTEGER PRIMARY KEY"
        if autoincrement:
            pk += " AUTOINCREMENT"
        self._cols_def.append(pk)
        self._pk_defined = True
        return self

    # Create a TEXT column
    def textCol(self, name: str, constraints: str = ""):
        if self.mode == "modtable" and self._col_to_modify:
            cd = self._buildColumnDef(name, "TEXT", constraints)
            self._modifyColumn(self._col_to_modify, cd)
            self._col_to_modify = None
            return self
        if self.mode in ["newtable", "addcolumns"]:
            cd = self._buildColumnDef(name, "TEXT", constraints)
            self._cols_def.append(cd)
        return self

    # Create an INTEGER column
    def intCol(self, name: str, constraints: str = ""):
        if self.mode == "modtable" and self._col_to_modify:
            cd = self._buildColumnDef(name, "INTEGER", constraints)
            self._modifyColumn(self._col_to_modify, cd)
            self._col_to_modify = None
            return self
        if self.mode in ["newtable", "addcolumns"]:
            cd = self._buildColumnDef(name, "INTEGER", constraints)
            self._cols_def.append(cd)
        return self

    # Create a REAL column
    def floatCol(self, name: str, constraints: str = ""):
        if self.mode == "modtable" and self._col_to_modify:
            cd = self._buildColumnDef(name, "REAL", constraints)
            self._modifyColumn(self._col_to_modify, cd)
            self._col_to_modify = None
            return self
        if self.mode in ["newtable", "addcolumns"]:
            cd = self._buildColumnDef(name, "REAL", constraints)
            self._cols_def.append(cd)
        return self

    # Create a DATE column
    def dateCol(self, name: str, constraints: str = ""):
        if self.mode == "modtable" and self._col_to_modify:
            cd = self._buildColumnDef(name, "DATE", constraints)
            self._modifyColumn(self._col_to_modify, cd)
            self._col_to_modify = None
            return self
        if self.mode in ["newtable", "addcolumns"]:
            cd = self._buildColumnDef(name, "DATE", constraints)
            self._cols_def.append(cd)
        return self

    # Create a foreign key
    def FK(self, column_name: str, ref_table: str, ref_pk: str = "id"):
        if self.mode == "modtable":
            if self._col_to_modify:
                self._col_to_modify = None
            # Not supported to add an FK in modtable
            raise NotImplementedError("Cannot add a foreign key in 'modtable' mode.")
        col_def = f"{column_name} INTEGER"
        self._cols_def.append(col_def)
        fk_def = f"FOREIGN KEY ({column_name}) REFERENCES {ref_table}({ref_pk})"
        self._fks_def.append(fk_def)
        return self

    # Build the table
    def create(self):
        if self.mode != "newtable":
            raise ValueError("create() can only be used in 'newtable' mode.")
        all_defs = self._cols_def + self._fks_def
        defs_str = ", ".join(all_defs)
        sql = f"CREATE TABLE IF NOT EXISTS {self.table_name} ( {defs_str} );"
        try:
            c = self.connection.cursor()
            c.execute(sql)
            self.connection.commit()
            print(f"[SUCCESS] Table '{self.table_name}' created or already exists.")
        except sqlite3.Error as e:
            print(f"[ERROR] Failed to create table '{self.table_name}': {e}")
        return self

    # Switch to addcolumns mode
    def add(self):
        if self.mode != "addcolumns":
            raise ValueError("add() can only be used in 'addcolumns' mode.")
        c = self.connection.cursor()
        for col_def in self._cols_def:
            try:
                sql = f"ALTER TABLE {self.table_name} ADD COLUMN {col_def};"
                c.execute(sql)
                self.connection.commit()
                print(f"[SUCCESS] Column '{col_def}' has been added to '{self.table_name}'.")
            except sqlite3.Error as e:
                print(f"[ERROR] Failed to add column '{col_def}' to '{self.table_name}': {e}")
        if len(self._fks_def) > 0:
            print("[WARNING] Adding a foreign key after table creation hcan corrupt references to the table in any existing triggers, views, and foreign key constraints.")
            self._addFK()
        return self

    # Switch to modtable mode
    def modTable(self, table_name: str):
        self.table_name = table_name
        self.mode = "modtable"
        return self

    # Rename the table
    def modName(self, new_name: str):
        if self.mode != "modtable":
            raise ValueError("modName() can only be used in 'modtable' mode.")
        try:
            c = self.connection.cursor()
            sql = f"ALTER TABLE {self.table_name} RENAME TO {new_name};"
            c.execute(sql)
            self.connection.commit()
            print(f"[SUCCESS] Table '{self.table_name}' was renamed to '{new_name}'.")
            self.table_name = new_name
        except sqlite3.Error as e:
            print(f"[ERROR] Failed to rename table '{self.table_name}' to '{new_name}': {e}")
        return self

    # Start column modify
    def modCol(self, old_col_name: str):
        if self.mode != "modtable":
            raise ValueError("modCol() can only be used in 'modtable' mode.")
        self._col_to_modify = old_col_name
        return self

    # Remove a column
    def remCol(self, column_name: str):
        if self.mode != "modtable":
            raise ValueError("remCol() can only be used in 'modtable' mode.")
        try:
            info = self._getTableInfo()
            new_schema = []
            for col in info:
                cid, cname, ctype, notnull, dflt, pk = col
                if cname == column_name:
                    continue
                base = f"{cname} {ctype}"
                if notnull:
                    base += " NOT NULL"
                if dflt is not None:
                    base += f" DEFAULT {dflt}"
                if pk:
                    base += " PRIMARY KEY"
                new_schema.append(base)
            temp = f"{self.table_name}_temp_remove"
            c = self.connection.cursor()
            create_temp = f"CREATE TABLE {temp} ( {', '.join(new_schema)} );"
            c.execute(create_temp)
            orig_cols = [x[1] for x in info if x[1] != column_name]
            new_cols = [x.split()[0] for x in new_schema]
            copy_sql = f"INSERT INTO {temp} ({', '.join(new_cols)}) SELECT {', '.join(orig_cols)} FROM {self.table_name};"
            c.execute(copy_sql)
            drop_sql = f"DROP TABLE {self.table_name};"
            c.execute(drop_sql)
            rename_sql = f"ALTER TABLE {temp} RENAME TO {self.table_name};"
            c.execute(rename_sql)
            self.connection.commit()
            print(f"[SUCCESS] Column '{column_name}' has been removed from '{self.table_name}'.")
        except sqlite3.Error as e:
            print(f"[ERROR] Failed to remove column '{column_name}' from '{self.table_name}': {e}")
        return self

    # Internal method to modify a column
    def _modifyColumn(self, old_col_name: str, new_def: str):
        try:
            info = self._getTableInfo()
            new_schema = []
            new_col_name = new_def.split()[0]
            for col in info:
                cid, cname, ctype, notnull, dflt, pk = col
                if cname == old_col_name:
                    new_schema.append(new_def)
                else:
                    base = f"{cname} {ctype}"
                    if notnull:
                        base += " NOT NULL"
                    if dflt is not None:
                        base += f" DEFAULT {dflt}"
                    if pk:
                        base += " PRIMARY KEY"
                    new_schema.append(base)
            temp = f"{self.table_name}_temp_alter"
            c = self.connection.cursor()
            create_temp = f"CREATE TABLE {temp} ( {', '.join(new_schema)} );"
            c.execute(create_temp)
            orig_cols = [x[1] for x in info]
            new_cols = [x.split()[0] for x in new_schema]
            old_list = []
            new_list = []
            for oc in orig_cols:
                if oc == old_col_name:
                    if new_col_name in new_cols:
                        old_list.append(oc)
                        new_list.append(new_col_name)
                else:
                    if oc in new_cols:
                        old_list.append(oc)
                        new_list.append(oc)
            copy_sql = f"INSERT INTO {temp} ({', '.join(new_list)}) SELECT {', '.join(old_list)} FROM {self.table_name};"
            c.execute(copy_sql)
            drop_sql = f"DROP TABLE {self.table_name};"
            c.execute(drop_sql)
            rename_sql = f"ALTER TABLE {temp} RENAME TO {self.table_name};"
            c.execute(rename_sql)
            self.connection.commit()
            print(f"[SUCCESS] Column '{old_col_name}' was modified to '{new_def}' in '{self.table_name}'.")
        except sqlite3.Error as e:
            print(f"[ERROR] Failed to modify column '{old_col_name}' in '{self.table_name}': {e}")

    # Internal method to add FK(s) to anexisting table
    def _addFK(self):
        # Connect to the SQLite database
        cursor = self.connection.cursor()
        table_name = self.table_name
        new_fk_definitions = self._fks_def
        try:
            # Get existing table schema
            cursor.execute(f"PRAGMA table_info({table_name})")
            table_info = cursor.fetchall()
            # Extract primary key and column definitions
            existing_columns_definitions = []
            existing_pk_columns = []
            for column in table_info:
                col_name = column[1]
                col_type = column[2]
                not_null = "NOT NULL" if column[3] else ""
                default_val = f"DEFAULT {column[4]}" if column[4] is not None else ""
                pk = column[5]  # Primary key flag
                # Build column definition
                column_definition = f"{col_name} {col_type} {not_null} {default_val}".strip()
                existing_columns_definitions.append(column_definition)
                # Track primary key columns
                if pk:
                    existing_pk_columns.append(col_name)
            # Build the primary key definition
            if existing_pk_columns:
                existing_pk_definition = f"PRIMARY KEY ({', '.join(existing_pk_columns)})"
            else:
                existing_pk_definition = ""
            # Combine all column definitions
            existing_columns_definitions_sql = ",\n    ".join(existing_columns_definitions)
            # Combine existing and new foreign key definitions
            new_fks_definitions_sql = ",\n    ".join(new_fk_definitions)
            # Assemble the CREATE TABLE statement
            create_table_sql = f"""
            CREATE TABLE {table_name}_new (
                {existing_columns_definitions_sql},
                {existing_pk_definition},
                {new_fks_definitions_sql}
            );
            """
            # Remove empty lines from SQL
            create_table_sql = "\n".join([line.strip() for line in create_table_sql.splitlines() if line.strip()])
            # Begin transaction and execute the schema modification
            cursor.execute("PRAGMA foreign_keys = OFF;")
            cursor.execute("BEGIN TRANSACTION;")
            cursor.execute(create_table_sql)
            cursor.execute(f"INSERT INTO {table_name}_new SELECT * FROM {table_name};")
            cursor.execute(f"DROP TABLE {table_name};")
            cursor.execute(f"ALTER TABLE {table_name}_new RENAME TO {table_name};")
            cursor.execute("COMMIT;")
            cursor.execute("PRAGMA foreign_keys = ON;")
            print(f"[SUCCESS] FK(s) successfully added to table '{table_name}'.")

        except sqlite3.Error as e:
            # Roll back in case of an error
            self.connection.rollback()
            print(f"[ERROR] Failed to add FK(s) to '{self.table_name}': {e}")


    # Internal method to get table info
    def _getTableInfo(self):
        c = self.connection.cursor()
        c.execute(f"PRAGMA table_info({self.table_name});")
        return c.fetchall()

    # Internal method to build column definition
    def _buildColumnDef(self, name: str, ctype: str, constraints: str):
        parts = []
        upper_con = constraints.upper()
        if "NN" in upper_con or "NOT NULL" in upper_con:
            parts.append("NOT NULL")
        if "UQ" in upper_con or "UNIQUE" in upper_con:
            parts.append("UNIQUE")
        joined = " ".join(parts)
        return f"{name} {ctype} {joined}".strip()



# EasyLiteRecord.py
import sqlite3
from typing import Any, List, Dict

class EasyLiteRecord:
    def __init__(self, core, table_name: str, mode: str):
        self.core = core
        self.connection = core.connection
        self.table_name = table_name
        self.mode = mode
        self._values_dict: Dict[str, Any] = {}
        self._where_clause: str = ""
        self._where_params: List[Any] = []
        self._table_info: List[Any] = []
        self._multi_rows: List[List[Any]] = []
        self._load_table_info()

    def _load_table_info(self):
        try:
            c = self.connection.cursor()
            c.execute(f"PRAGMA table_info({self.table_name});")
            self._table_info = c.fetchall()
        except sqlite3.Error as e:
            print(f"[ERROR] Failed to load table info for '{self.table_name}': {e}")
            self._table_info = []

    def row(self, *values: Any):
        if self.mode not in ("insert", "update"):
            print(f"[ERROR] .row(...) can only be used in 'insert' or 'update' mode (current: {self.mode}).")
            return self

        # insert mode -> accumulate in _multi_rows
        if self.mode == "insert":
            self._multi_rows.append(values)
            return self

        # update mode -> map row values into _values_dict for non-PK columns
        if self.mode == "update":
            non_pk_cols = [col for col in self._table_info if col[5] == 0]
            col_names = [c[1] for c in non_pk_cols]
            usable_count = min(len(values), len(col_names))

            for i in range(usable_count):
                val = values[i]
                col_name = col_names[i]
                if val is self.core.skip:
                    continue
                elif val is self.core.null:
                    self._values_dict[col_name] = None
                else:
                    self._values_dict[col_name] = val

            return self

    def multiRows(self, rows: List[List[Any]]):
        if self.mode != "insert":
            print("[ERROR] multiRows(...) can only be used in 'insert' mode.")
            return self
        self._multi_rows.extend(rows)
        return self

    def field(self, column_name: str, value: Any):
        # if user provides db.skip, do nothing
        if value is self.core.skip:
            return self

        # if user provides db.null, store None so SQLite can handle it
        if value is self.core.null:
            self._values_dict[column_name] = None
        else:
            self._values_dict[column_name] = value

        return self

    def where(self, clause: str, *params):
        self._where_clause = clause
        self._where_params = list(params)
        return self

    def record(self):
        if self.mode == "insert":
            if self._multi_rows:
                return self._insert_multi()
            else:
                return self._insert_single()
        elif self.mode == "update":
            return self._update_record()
        elif self.mode == "delete":
            print("[ERROR] 'record()' not valid for delete mode. Use 'execute()'.")
        else:
            print("[ERROR] Unknown mode.")
        return self

    def execute(self):
        if self.mode != "delete":
            print("[ERROR] execute() is for delete mode only.")
            return self
        try:
            sql = f"DELETE FROM {self.table_name}"
            if self._where_clause:
                sql += f" WHERE {self._where_clause}"
            c = self.connection.cursor()
            c.execute(sql, self._where_params)
            self.connection.commit()
            print(f"[SUCCESS] Records deleted from '{self.table_name}'.")
        except sqlite3.Error as e:
            print(f"[ERROR] {e}")
        return self

    def _insert_single(self):
        if not self._values_dict:
            print("[WARNING] No fields set for insert, and no rows queued. Nothing inserted.")
            return self

        cols = list(self._values_dict.keys())
        placeholders = ", ".join("?" for _ in cols)
        vals = [self._values_dict[c] for c in cols]

        sql = f"INSERT INTO {self.table_name} ({', '.join(cols)}) VALUES ({placeholders})"
        try:
            c = self.connection.cursor()
            c.execute(sql, vals)
            self.connection.commit()
            print(f"[SUCCESS] Inserted a new record into '{self.table_name}'.")
        except sqlite3.Error as e:
            print(f"[ERROR] {e}")
        return self

    def _insert_multi(self):
        if not self._table_info:
            print("[ERROR] Table info not loaded. Multi insert impossible.")
            return self

        non_pk = [col for col in self._table_info if col[5] == 0]
        col_names = [c[1] for c in non_pk]
        placeholders = ", ".join("?" for _ in col_names)
        sql = f"INSERT INTO {self.table_name} ({', '.join(col_names)}) VALUES ({placeholders})"

        c = self.connection.cursor()
        inserted_count = 0

        try:
            for row_vals in self._multi_rows:
                final_vals = []
                for i in range(len(col_names)):
                    if i < len(row_vals):
                        v = row_vals[i]
                        if v is self.core.skip:
                            final_vals.append(None)
                        elif v is self.core.null:
                            final_vals.append(None)
                        else:
                            final_vals.append(v)
                    else:
                        final_vals.append(None)
                c.execute(sql, final_vals)
                inserted_count += 1
            self.connection.commit()
            print(f"[SUCCESS] Inserted {inserted_count} records into '{self.table_name}'.")
        except sqlite3.Error as e:
            print(f"[ERROR] {e}")
        return self

    def _update_record(self):
        if not self._values_dict:
            print("[WARNING] No fields set for update. Nothing will be updated.")
            return self

        set_clause = ", ".join(f"{col} = ?" for col in self._values_dict.keys())
        vals = list(self._values_dict.values())

        sql = f"UPDATE {self.table_name} SET {set_clause}"

        if self._where_clause:
            sql += f" WHERE {self._where_clause}"
            vals += self._where_params
        else:
            print("[WARNING] No WHERE clause specified. Updating ALL rows.")

        try:
            c = self.connection.cursor()
            c.execute(sql, vals)
            self.connection.commit()
            print(f"[SUCCESS] Updated records in '{self.table_name}'.")
        except sqlite3.Error as e:
            print(f"[ERROR] {e}")

        return self



# EasyLiteCore.py
import os
import sqlite3
from .EasyLiteBuild import EasyLiteBuild
from .EasyLiteQuery import EasyLiteQuery
from .EasyLiteRecord import EasyLiteRecord
from .EasyLiteResult import EasyLiteResult


# Main class for database operations
class EasyLiteCore:
    # Constructor
    def __init__(self):
        self.connection = None
        self.cursor = None
        self.db_path = None
        self.skip=object()
        self.null=object()

    # Connect to or create a SQLite database
    def connect(self, db_path: str):
        try:
            db_exists = os.path.exists(db_path)
            self.connection = sqlite3.connect(db_path)
            self.cursor = self.connection.cursor()
            self.db_path = db_path
            if db_exists:
                print(f"[SUCCESS] Successfully connected to existing database: '{db_path}'.")
            else:
                print(f"[SUCCESS] Database file did not exist. A new database has been created: '{db_path}'.")
        except sqlite3.Error as e:
            print(f"[ERROR] Could not connect to the database: {e}")
        return self

    # Returns a builder for creating a new table
    def newTable(self, table_name: str) -> EasyLiteBuild:
        return EasyLiteBuild(self.connection, table_name, mode="newtable")

    # Returns a builder for adding columns to an existing table
    def addToTable(self, table_name: str) -> EasyLiteBuild:
        return EasyLiteBuild(self.connection, table_name, mode="addcolumns")

    # Returns a builder for modifying an existing table
    def modTable(self, table_name: str) -> EasyLiteBuild:
        return EasyLiteBuild(self.connection, table_name, mode="modtable")

    # Insert resord(s)
    def insertIn(self, table_name: str):
        return EasyLiteRecord(self, table_name, mode="insert")

    # Update resord(s)
    def updateIn(self, table_name: str):
        return EasyLiteRecord(self, table_name, mode="update")

    # Delete resord(s)
    def deleteIn(self, table_name: str):
        return EasyLiteRecord(self, table_name, mode="delete")

    # Builds a SELECT query
    def select(self, table_name: str) -> EasyLiteQuery:
        return EasyLiteQuery(self.connection, table_name)

    # Execute a custom SQL query
    def executeCustomQuery(self, sql: str, params: tuple = ()) -> EasyLiteResult:
        try:
            c = self.connection.cursor()
            c.execute(sql, params)
            rows = []
            col_names = []
            try:
                rows = c.fetchall()
                col_names = [desc[0] for desc in c.description]
            except sqlite3.ProgrammingError:
                pass
            self.connection.commit()
            return EasyLiteResult(rows, col_names)
        except sqlite3.Error as e:
            print(f"[ERROR] Custom query failed: {e}")
            return EasyLiteResult([], [])

    # Drop an existing table
    def dropTable(self, table_name: str):
        q = f"DROP TABLE IF EXISTS {table_name};"
        try:
            self.cursor.execute(q)
            self.connection.commit()
            print(f"[SUCCESS] Successfully dropped table '{table_name}' (if it existed).")
        except sqlite3.Error as e:
            print(f"[ERROR] Failed to drop table '{table_name}': {e}")

    # Print a formatted database schema
    def getSchema(self, table_name=None):
        try:
            c = self.connection.cursor()
            if table_name:
                tables = [(table_name,)]
            else:
                c.execute("SELECT name FROM sqlite_master WHERE type='table' AND name NOT LIKE 'sqlite_%' ORDER BY name;")
                tables = c.fetchall()
            if not tables:
                print("[ERROR] No user-defined tables found in the database.")
                return
            print("=== DATABASE SCHEMA ===\n")
            for i, (tname,) in enumerate(tables, start=1):
                print(f"{i}) Table: {tname}")
                try:
                    c.execute(f"PRAGMA table_info({tname});")
                    cols = c.fetchall()
                    print("   Columns:")
                    for col in cols:
                        cid, name, ctype, notnull, dflt, pk = col
                        nn = " NOT NULL" if notnull else ""
                        pkstr = " [PK]" if pk else ""
                        df = f" DEFAULT {dflt}" if dflt is not None else ""
                        print(f"     - {name} {ctype}{nn}{df}{pkstr}")

                    c.execute(f"PRAGMA foreign_key_list({tname});")
                    fks = c.fetchall()
                    if fks:
                        print("   Foreign Keys:")
                        for fk in fks:
                            _, _, rtab, col_local, col_remote, on_up, on_del, _ = fk
                            print(f"     - {col_local} -> {rtab}({col_remote}) (ON UPDATE {on_up}, ON DELETE {on_del})")
                    else:
                        print("   Foreign Keys: none")
                except sqlite3.Error as e:
                    print(f"[ERROR] Could not retrieve schema info for table '{tname}': {e}")
                print("")
        except sqlite3.Error as e:
            print(f"[ERROR] Could not retrieve database schema: {e}")

    # Delete database file
    def deleteDatabaseFile(self):
        self.close()
        os.remove(self.db_path)

    # Close the database connection
    def close(self):
        if self.connection:
            try:
                self.connection.close()
                self.connection = None
                print("[SUCCESS] Database connection has been closed.")
            except sqlite3.Error as e:
                print(f"[ERROR] Failed to close the database connection: {e}")




# EasyLiteResult.py
import csv
import json
from typing import List, Tuple, Any

# Class to handle query results
class EasyLiteResult:
    # Constructor
    def __init__(self, rows: List[Tuple[Any]], columns: List[str]):
        self._rows = rows
        self._columns = columns

    # Returns all rows
    def rows(self) -> List[Tuple[Any]]:
        return self._rows

    # Returns column names
    def columns(self) -> List[str]:
        return self._columns

    # Returns the total number of rows
    def count(self) -> int:
        return len(self._rows)

    # Returns a list of dicts, each dict representing a row
    def toDict(self) -> List[dict]:
        data = []
        for row in self._rows:
            item = {}
            for col, val in zip(self._columns, row):
                item[col] = val
            data.append(item)
        return data

    # Returns the CSV content as a string
    def toCSV(self) -> str:
        try:
            import io
            buffer = io.StringIO()
            writer = csv.writer(buffer)
            writer.writerow(self._columns)
            writer.writerows(self._rows)
            return buffer.getvalue()
        except Exception as e:
            print(f"[ERROR] Failed to generate CSV string: {e}")
            return ""

    # Exports the result to a CSV file
    def exportCSV(self, csv_filename: str):
        try:
            with open(csv_filename, "w", newline="", encoding="utf-8") as f:
                writer = csv.writer(f)
                writer.writerow(self._columns)
                writer.writerows(self._rows)
            print(f"[SUCCESS] CSV file has been successfully exported to '{csv_filename}'.")
        except Exception as e:
            print(f"[ERROR] Failed to export CSV file '{csv_filename}': {e}")

    # Shorter, nicer name for printing results
    def show(self):
        if not self._columns:
            print("(No columns in result)")
            return
        col_widths = []
        for i, col_name in enumerate(self._columns):
            try:
                max_len_in_col = max((len(str(row[i])) for row in self._rows), default=0)
            except Exception:
                max_len_in_col = 0
            col_widths.append(max(len(col_name), max_len_in_col))
        header_line = " | ".join(col_name.ljust(col_widths[i]) for i, col_name in enumerate(self._columns))
        print(header_line)
        print("-" * len(header_line))
        for row in self._rows:
            row_line = " | ".join(str(row[i]).ljust(col_widths[i]) for i in range(len(self._columns)))
            print(row_line)

    # Returns the result in a standard JSON array
    def toJSON(self) -> str:
        try:
            return json.dumps(self.toDict())
        except Exception as e:
            print(f"[ERROR] Failed to generate JSON: {e}")
            return "[]"

    # Exports the result as a JSON file with minimal wrapping
    def exportJSON(self, json_filename: str):
        try:
            data_str = self.toJSON()
            with open(json_filename, "w", encoding="utf-8") as f:
                f.write(data_str)
            print(f"[SUCCESS] JSON file has been successfully exported to '{json_filename}'.")
        except Exception as e:
            print(f"[ERROR] Failed to export JSON file '{json_filename}': {e}")

    # Returns the result in an "API style" JSON, e.g. with status, count, data
    def toApiJSON(self) -> str:
        try:
            api_data = {
                "status": "success",
                "count": self.count(),
                "data": self.toDict()
            }
            return json.dumps(api_data)
        except Exception as e:
            print(f"[ERROR] Failed to generate API JSON: {e}")
            return '{"status":"error","count":0,"data":[]}'
